/***
 * Cecilia ADL Compiler
 * Copyright (C) 2006-2008 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@objectweb.org
 *
 * Author: Matthieu Leclercq
 */

package org.objectweb.fractal.cecilia.adl;

import java.io.File;
import java.io.PrintStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Abstract Cecilia compiler Launcher.
 */
public abstract class AbstractLauncher {

  protected final CmdPathOption srcPathOpt = new CmdPathOption(
                                               "S",
                                               "src-path",
                                               "the search path of ADL,IDL and implementation files (list of path separated by '"
                                                   + File.pathSeparator + "')",
                                               "<path list>");

  protected final CmdArgument   outDirOpt  = new CmdArgument(
                                               "o",
                                               "out-path",
                                               "the path where generated files will be put",
                                               "<output path>", ".", false);

  protected final CmdFlag       helpOpt    = new CmdFlag("h", "help",
                                               "Print this help and exit");

  protected final Options       options    = new Options();
  {
    options.addOptions(helpOpt, srcPathOpt, outDirOpt);
  }

  // ---------------------------------------------------------------------------
  // Utility methods
  // ---------------------------------------------------------------------------

  protected ClassLoader getSourceClassLoader(final List<String> srcPath) {
    final List<String> validatedPaths = new ArrayList<String>(srcPath.size());

    // check source paths
    for (final String path : srcPath) {
      final File f = new File(path);
      if (!f.exists()) {
        System.out.println("Warning '" + f.getAbsolutePath()
            + "' source path can't be found ");
      } else if (!f.isDirectory()) {
        System.out.println("Warning: \"" + path
            + "\" is not a directory, path ignored.");
      } else {
        validatedPaths.add(path);
      }
    }

    // build URL array of source path
    final URL[] urls = new URL[validatedPaths.size()];
    for (int i = 0; i < urls.length; i++) {
      final String path = validatedPaths.get(i);
      final File f = new File(path);
      try {
        urls[i] = f.toURI().toURL();
      } catch (final MalformedURLException e) {
        // never append
        throw new Error(e);
      }
    }

    return new URLClassLoader(urls, getClass().getClassLoader());
  }

  protected ClassLoader getSourceClassLoader(
      final Map<String, List<String>> argMap) {
    final List<String> srcPath = new ArrayList<String>();
    List<String> args = argMap.get(srcPathOpt.getShortName());
    if (args != null) {
      for (final String arg : args) {
        srcPath.addAll(parsePathList(arg));
      }
    }

    args = argMap.get(srcPathOpt.getLongName());
    if (args != null) {
      for (final String arg : args) {
        srcPath.addAll(parsePathList(arg));
      }
    }

    if (srcPath.isEmpty()) {
      srcPath.add(".");
    }

    return getSourceClassLoader(srcPath);
  }

  protected ClassLoader getSourceClassLoader(final CommandLine cmdLine) {
    List<String> srcPath = srcPathOpt.getPathValue(cmdLine);
    if (srcPath == null) {
      srcPath = new ArrayList<String>(1);
      srcPath.add(".");
    }

    return getSourceClassLoader(srcPath);
  }

  protected static File newBuildDir(final File outDir, final String dirName)
      throws InvalidCommandLineException {
    final File d = new File(outDir, dirName);
    checkDir(d);
    return d;
  }

  protected static File newBuildDir(final String name)
      throws InvalidCommandLineException {
    final File d = new File(name);
    checkDir(d);
    return d;
  }

  protected static void checkDir(final File d)
      throws InvalidCommandLineException {
    if (d.exists() && !d.isDirectory())
      throw new InvalidCommandLineException("Invalid build directory '"
          + d.getAbsolutePath() + "' not a directory", 6);
  }

  protected static List<String> parsePathList(String paths) {
    final List<String> l = new ArrayList<String>();
    int index = paths.indexOf(File.pathSeparatorChar);
    while (index != -1) {
      l.add(paths.substring(0, index));
      paths = paths.substring(index + 1);
      index = paths.indexOf(File.pathSeparatorChar);
    }
    l.add(paths);
    return l;
  }

  protected static CommandLine argsToCommandLine(final Options options,
      final boolean allowUnknownOption, final String... args)
      throws InvalidCommandLineException {
    final CommandLine cmdLine = new CommandLine(options);

    for (final String arg : args) {
      if (arg.startsWith("-")) {

        final String argName;
        final String argValue;

        boolean longName;
        final int startIndex;
        if (arg.startsWith("--")) {
          startIndex = 2;
          longName = true;
        } else {
          startIndex = 1;
          longName = false;
        }

        final int index = arg.indexOf('=');
        if (index == -1) {
          argName = arg.substring(startIndex);
          argValue = null;
        } else {
          if (index < startIndex + 1) {
            throw new InvalidCommandLineException("Invalid option '" + arg
                + "'", 1);
          }
          argName = arg.substring(startIndex, index);
          argValue = arg.substring(index + 1);
        }

        final CmdOption opt;
        if (longName)
          opt = cmdLine.options.getByLongName(argName);
        else
          opt = cmdLine.options.getByName(argName);

        if (opt == null) {
          if (allowUnknownOption) {
            cmdLine.addArgument(arg);
          } else {
            throw new InvalidCommandLineException("Unknown option '" + argName
                + "'", 1);
          }
        } else {
          if (opt instanceof CmdFlag) {
            if (argValue != null) {
              throw new InvalidCommandLineException("Invalid option '"
                  + argName + "' do not accept value", 1);
            }

            ((CmdFlag) opt).setPresent(cmdLine);
          } else { // opt instanceof CmdArgument
            if (argValue == null) {
              throw new InvalidCommandLineException("Invalid option '"
                  + argName + "' expects a value", 1);
            }
            ((CmdArgument) opt).setValue(cmdLine, argValue);
          }
        }
      } else {
        cmdLine.addArgument(arg);
      }
    }
    return cmdLine;
  }

  protected static Map<String, List<String>> argsToMap(final String... args)
      throws InvalidCommandLineException {
    final Map<String, List<String>> map = new LinkedHashMap<String, List<String>>();
    List<String> nonOptList = null;

    for (final String arg : args) {
      if (arg.startsWith("-")) {

        final String argName;
        final String argValue;

        final int startIndex;
        if (arg.startsWith("--"))
          startIndex = 2;
        else
          startIndex = 1;

        final int index = arg.indexOf('=');
        if (index == -1) {
          argName = arg.substring(1);
          argValue = null;
        } else {
          if (index < startIndex + 1) {
            throw new InvalidCommandLineException("Invalid option '" + arg
                + "'", 1);
          }
          argName = arg.substring(startIndex, index);
          argValue = arg.substring(index + 1);
        }

        List<String> l = map.get(argName);
        if (l == null) {
          l = new ArrayList<String>();
          map.put(argName, l);
        }
        l.add(argValue);

      } else {
        if (nonOptList == null) {
          nonOptList = new ArrayList<String>();
          map.put(null, nonOptList);
        }
        nonOptList.add(arg);
      }
    }

    return map;
  }

  protected abstract void printUsage(PrintStream ps);

  protected void printHelp(final PrintStream ps) {
    printUsage(ps);
    ps.println();
    ps.println("Available options are :");
    int maxCol = 0;

    for (final CmdOption opt : options.getOptions()) {
      final int col = 2 + opt.getPrototype().length();
      if (col > maxCol) maxCol = col;
    }
    for (final CmdOption opt : options.getOptions()) {
      final StringBuffer sb = new StringBuffer("  ");
      sb.append(opt.getPrototype());
      while (sb.length() < maxCol)
        sb.append(' ');
      sb.append("  ").append(opt.getDescription());
      ps.println(sb);
    }
  }

  // ---------------------------------------------------------------------------
  // Internal classes
  // ---------------------------------------------------------------------------

  /**
   * Exception thrown when an error on the command line has been detected.
   */
  public static class InvalidCommandLineException extends Exception {

    final int exitValue;

    /**
     * @param message detail message.
     * @param exitValue exit value.
     */
    public InvalidCommandLineException(final String message, final int exitValue) {
      super(message);
      this.exitValue = exitValue;
    }

  }

  protected static class Options {
    protected final Set<CmdOption>         optionSet          = new LinkedHashSet<CmdOption>();
    protected final Map<String, CmdOption> optionsByShortName = new HashMap<String, CmdOption>();
    protected final Map<String, CmdOption> optionsByLongName  = new HashMap<String, CmdOption>();

    public void addOption(final CmdOption option) {
      if (option.shortName != null) {
        final CmdOption prevOpt = optionsByShortName.put(option.shortName,
            option);
        if (prevOpt != null) {
          throw new IllegalArgumentException("short name '" + option.shortName
              + "' already used");
        }
      }
      if (option.longName != null) {
        final CmdOption prevOpt = optionsByLongName
            .put(option.longName, option);
        if (prevOpt != null) {
          throw new IllegalArgumentException("long name '" + option.longName
              + "' already used");
        }
      }

      optionSet.add(option);
    }

    public void addOptions(final CmdOption... options) {
      for (final CmdOption option : options) {
        addOption(option);
      }
    }

    public Collection<CmdOption> getOptions() {
      return optionSet;
    }

    public CmdOption getByShortName(final String shortName) {
      return optionsByShortName.get(shortName);
    }

    public CmdOption getByLongName(final String longName) {
      return optionsByLongName.get(longName);
    }

    CmdOption getByName(final String name) {
      final CmdOption option = optionsByShortName.get(name);
      if (option == null)
        return optionsByLongName.get(name);
      else
        return option;
    }
  }

  protected static class CommandLine {
    protected final Options                options;
    protected final Map<CmdOption, String> optionValues = new LinkedHashMap<CmdOption, String>();
    protected final List<String>           arguments    = new ArrayList<String>();

    public CommandLine(final Options options) {
      this.options = options;
    }

    protected void addArgument(final String argument) {
      arguments.add(argument);
    }

    protected String setOptionValue(final CmdOption option, final String value) {
      return optionValues.put(option, value);
    }

    public List<String> getArguments() {
      return arguments;
    }

    protected boolean isOptionPresent(final CmdOption option) {
      return optionValues.containsKey(option);
    }

    protected String getOptionValue(final String name) {
      for (final Map.Entry<CmdOption, String> optionValue : optionValues
          .entrySet()) {
        if (name.equals(optionValue.getKey().shortName)
            || name.equals(optionValue.getKey().longName))
          return optionValue.getValue();
      }

      final CmdOption opt = options.getByName(name);
      if (opt instanceof CmdArgument) return ((CmdArgument) opt).defaultValue;
      return null;
    }

    protected String getOptionValue(final CmdOption option) {
      final String value = optionValues.get(option);
      return value;
    }
  }

  protected abstract static class CmdOption {
    protected final String shortName;
    protected final String longName;
    protected final String description;

    public CmdOption(final String shortName, final String longName,
        final String description) {
      if (shortName == null && longName == null) {
        throw new IllegalArgumentException("Invalid option names");
      }
      if (shortName != null && shortName.length() > 1) {
        throw new IllegalArgumentException("Invalid shortName");
      }
      this.shortName = shortName;
      this.longName = longName;
      this.description = description;
    }

    public String getPrototype() {
      String desc;
      if (shortName != null) {
        desc = "-" + shortName;
        if (longName != null) {
          desc += ", --" + longName;
        }
      } else {
        desc = "--" + longName;
      }
      return desc;
    }

    public String getShortName() {
      return shortName;
    }

    public String getLongName() {
      return longName;
    }

    public String getDescription() {
      return description;
    }

    @Override
    public boolean equals(final Object obj) {
      if (this == obj) return true;

      if (!(obj instanceof CmdOption)) return false;

      final CmdOption opt = (CmdOption) obj;
      if (shortName == null) {
        return opt.shortName == null && opt.longName.equals(longName);
      } else {
        return shortName.equals(opt.shortName);
      }
    }

    @Override
    public int hashCode() {
      if (shortName == null)
        return longName.hashCode();
      else
        return shortName.hashCode();
    }
  }

  protected static class CmdFlag extends CmdOption {

    public CmdFlag(final String shortName, final String longName,
        final String description) {
      super(shortName, longName, description);
    }

    protected void setPresent(final CommandLine commandLine) {
      commandLine.setOptionValue(this, "");
    }

    public boolean isPresent(final CommandLine commandLine) {
      return commandLine.isOptionPresent(this);
    }
  }

  protected static class CmdArgument extends CmdOption {
    protected final String  argDesc;
    protected final String  defaultValue;
    protected final boolean allowMultiple;

    public CmdArgument(final String shortName, final String longName,
        final String description, final String argDesc,
        final String defaultValue, final boolean allowMultiple) {
      super(shortName, longName, (defaultValue == null)
          ? description
          : description + " (default is '" + defaultValue + "')");
      this.argDesc = argDesc;
      this.defaultValue = defaultValue;
      this.allowMultiple = allowMultiple;
    }

    public CmdArgument(final String shortName, final String longName,
        final String description, final String argDesc) {
      this(shortName, longName, description, argDesc, null, false);
    }

    public void setValue(final CommandLine commandLine, final String value)
        throws InvalidCommandLineException {
      if (value == null) return;
      final Object prevValue = commandLine.setOptionValue(this, value);
      if (!allowMultiple && prevValue != null) {
        throw new InvalidCommandLineException("'" + longName
            + "' can't be specified several times.", 1);
      }
    }

    public String getValue(final CommandLine commandLine) {
      final String optionValue = commandLine.getOptionValue(this);
      return optionValue == null ? defaultValue : optionValue;
    }

    public String getDefaultValue() {
      return defaultValue;
    }

    @Override
    public String getPrototype() {
      String desc;
      if (shortName != null) {
        desc = "-" + shortName + "=" + argDesc;
        if (longName != null) {
          desc += ", --" + longName;
        }
      } else {
        desc = "--" + longName + "=" + argDesc;
      }
      return desc;
    }

    @Override
    public String getDescription() {
      if (allowMultiple)
        return super.getDescription()
            + ". This option may be specified several times.";
      else
        return super.getDescription();
    }
  }

  protected static class CmdAppendOption extends CmdArgument {

    protected final String separator;

    public CmdAppendOption(final String shortName, final String longName,
        final String description, final String argDesc) {
      this(shortName, longName, description, argDesc, null, " ");
    }

    public CmdAppendOption(final String shortName, final String longName,
        final String description, final String argDesc,
        final String defaultValue, final String separator) {
      super(shortName, longName, description, argDesc, defaultValue, true);
      this.separator = separator;
    }

    @Override
    public void setValue(final CommandLine commandLine, final String value)
        throws InvalidCommandLineException {
      if (value == null) return;
      final String prevValue = commandLine.getOptionValue(this);

      if (prevValue == null) {
        commandLine.setOptionValue(this, value);
      } else {
        commandLine.setOptionValue(this, prevValue + separator + value);
      }
    }

    @Override
    public String getValue(final CommandLine commandLine) {
      return commandLine.getOptionValue(this);
    }
  }

  protected static class CmdPathOption extends CmdAppendOption {

    public CmdPathOption(final String shortName, final String longName,
        final String description, final String argDesc) {
      super(shortName, longName, description, argDesc, null, File.pathSeparator);
    }

    public List<String> getPathValue(final CommandLine commandLine) {
      final String value = getValue(commandLine);
      if (value == null)
        return null;
      else
        return parsePathList(value);
    }
  }
}
